#include "CIndexer.h"
#include "CXIndexDataConsumer.h"
#include "CXTranslationUnit.h"
#include "clang/Frontend/CompilerInstance.h"
#include "clang/Frontend/FrontendAction.h"
#include "clang/Lex/PPConditionalDirectiveRecord.h"
#include "llvm/Support/CrashRecoveryContext.h"
using namespace clang;
        using namespace cxtu;
        using namespace cxindex;
        
       class PPRegion {
        llvm::sys::fs::UniqueID UniqueID;
        time_t ModTime;
        unsigned Offset;
      public:   
        PPRegion(llvm::sys::fs::UniqueID UniqueID, unsigned offset, time_t modTime)  ;
        friend bool operator==(const PPRegion &lhs, const PPRegion &rhs) {
         return lhs.UniqueID == rhs.UniqueID && lhs.Offset == rhs.Offset &&            lhs.ModTime == rhs.ModTime;
       }
      };
       typedef llvm::DenseSet<PPRegion> PPRegionSetTy;
       
         namespace llvm {
         template <>   struct DenseMapInfo<PPRegion> {
          static inline PPRegion getEmptyKey() {
           return PPRegion(llvm::sys::fs::UniqueID(), unsigned(), 0);
         }
          static inline PPRegion getTombstoneKey() {
           return PPRegion(llvm::sys::fs::UniqueID(), unsigned(), 0);
         }
          static unsigned getHashValue(const PPRegion &S) {
           llvm::FoldingSetNodeID ID;
           return ID.ComputeHash();
         }
          static bool isEqual(const PPRegion &LHS, const PPRegion &RHS) {
           return LHS == RHS;
         }
        };
       }
        
       class SessionSkipBodyData ;
       class TUSkipBodyControl {
        SessionSkipBodyData &SessionData;
        PPConditionalDirectiveRecord &PPRec;
        Preprocessor &PP;
        PPRegionSetTy ParsedRegions;
        bool LastIsParsed;
      public:   
        bool isParsed(SourceLocation Loc, FileID FID, const FileEntry *FE) {
         PPRegion region = getRegion(Loc, FID, FE);
         LastIsParsed = ParsedRegions.count(region);
       }
      private:   PPRegion getRegion(SourceLocation Loc, FileID FID, const FileEntry *FE) ;
      };
       class IndexPPCallbacks : public PPCallbacks {};
       class IndexingConsumer : public ASTConsumer {
        CXIndexDataConsumer &DataConsumer;
        TUSkipBodyControl *SKCtrl;
      public:   IndexingConsumer(CXIndexDataConsumer &dataConsumer, TUSkipBodyControl *skCtrl)     : DataConsumer(dataConsumer), SKCtrl() {}
        bool shouldSkipFunctionBody(Decl *D) override {
         const SourceManager &SM = DataConsumer.getASTContext().getSourceManager();
         SourceLocation Loc = D->getLocation();
          FileID FID;
         const FileEntry *FE = SM.getFileEntryForID(FID);
         return SKCtrl->isParsed(Loc, FID, FE);
       }
      };
       class CaptureDiagnosticConsumer : public DiagnosticConsumer {};
       class IndexingFrontendAction : public ASTFrontendAction {
        std::shared_ptr<CXIndexDataConsumer> DataConsumer;
        SessionSkipBodyData *SKData;
        std::unique_ptr<TUSkipBodyControl> SKCtrl;
      public:   IndexingFrontendAction(std::shared_ptr<CXIndexDataConsumer> dataConsumer,                          SessionSkipBodyData *skData)       : DataConsumer(), SKData() {}
        std::unique_ptr<ASTConsumer> CreateASTConsumer(CompilerInstance &CI,                                                  StringRef InFile) override {
         return std::make_unique<IndexingConsumer>(*DataConsumer, SKCtrl.get());
       }
      };
       struct IndexSessionData {
        CXIndex CIdx;
        std::unique_ptr<SessionSkipBodyData> SkipBodyData;
      };
       
         static CXErrorCode clang_indexSourceFile_Impl(     CXIndexAction cxIdxAction, CXClientData client_data,     IndexerCallbacks *client_index_callbacks, unsigned index_callbacks_size,     unsigned index_options, const char *source_filename,     const char *const *command_line_args, int num_command_line_args,     ArrayRef<CXUnsavedFile> unsaved_files, CXTranslationUnit *out_TU,     unsigned TU_options) {
         IndexerCallbacks CB;
         IndexSessionData *IdxSession = static_cast<IndexSessionData *>(cxIdxAction);
         CIndexer *CXXIdx = static_cast<CIndexer *>(IdxSession->CIdx);
         CaptureDiagsKind CaptureDiagnostics = CaptureDiagsKind::All;
         CaptureDiagnosticConsumer *CaptureDiag = nullptr;
         IntrusiveRefCntPtr<DiagnosticsEngine>     Diags;
         std::unique_ptr<std::vector<const char *>> Args(       new std::vector<const char *>);
         std::shared_ptr<CompilerInvocation> CInvok =       createInvocationFromCommandLine(*Args, Diags);
         auto Unit = ASTUnit::create(CInvok, Diags, CaptureDiagnostics,                               /*UserFilesAreVolatile=*/true);
         std::unique_ptr<CXTUOwner> CXTU(       new CXTUOwner(MakeCXTranslationUnit(CXXIdx, std::move(Unit))));
         bool SkipBodies = 0 &&       CInvok->getLangOpts()->CPlusPlus;
         auto DataConsumer =     std::make_shared<CXIndexDataConsumer>(client_data, CB, index_options,                                           CXTU->getTU());
         auto InterAction = std::make_unique<IndexingFrontendAction>(DataConsumer,                          SkipBodies ? IdxSession->SkipBodyData.get() : nullptr);
       }
        void  clang_indexSourceFileFullArgv(     CXIndexAction idxAction, CXClientData client_data,     IndexerCallbacks *index_callbacks,                        CXTranslationUnit *out_TU) {
         unsigned int TU_options = 0;
         unsigned int num_unsaved_files = 0;
         struct CXUnsavedFile * unsaved_files = 0;
         int num_command_line_args = 0;
         const char *const * command_line_args = 0;
         const char * source_filename = 0;
         unsigned int index_options = 0;
         unsigned int index_callbacks_size = 0;
         CXErrorCode result = CXError_Failure;
         auto IndexSourceFileImpl = [=, &result] {
          result = clang_indexSourceFile_Impl(         idxAction, client_data, index_callbacks, index_callbacks_size,         index_options, source_filename, command_line_args,         num_command_line_args,         llvm::makeArrayRef(unsaved_files, num_unsaved_files), out_TU,         TU_options);
        };
         llvm::CrashRecoveryContext CRC;
         if (!RunSafely(CRC, IndexSourceFileImpl)) ;
       }
